@@ -1,4 +1,4 @@
-use mlua::{UserDataMethods, Lua};
+use mlua::{Lua, UserDataMethods};
 
 fn main() {
     let lua = Lua::new();
@@ -6,9 +6,10 @@ fn main() {
     lua.register_userdata_type::<String>(|reg| {
         let s = String::new();
         let mut s = &s;
-        reg.add_async_method("t", |_, this: &String, ()| async {
-            s = this;
+        reg.add_async_method("t", |_, this, ()| async {
+            s = &*this;
             Ok(())
         });
-    }).unwrap();
+    })
+    .unwrap();
 }
