@@ -1,10 +1,9 @@
 #![cfg(not(feature = "luau"))]
 
-use std::ops::Deref;
 use std::sync::atomic::{AtomicI64, Ordering};
 use std::sync::{Arc, Mutex};
 
-use mlua::{DebugEvent, Error, HookTriggers, Lua, Result, Value};
+use mlua::{DebugEvent, Error, HookTriggers, Lua, Result, ThreadStatus, Value, VmState};
 
 #[test]
 fn test_hook_triggers() {
@@ -25,9 +24,9 @@ fn test_line_counts() -> Result<()> {
     let lua = Lua::new();
     lua.set_hook(HookTriggers::EVERY_LINE, move |_lua, debug| {
         assert_eq!(debug.event(), DebugEvent::Line);
-        hook_output.lock().unwrap().push(debug.curr_line());
-        Ok(())
-    });
+        hook_output.lock().unwrap().push(debug.current_line().unwrap());
+        Ok(VmState::Continue)
+    })?;
     lua.load(
         r#"
             local x = 2 + 3
@@ -61,8 +60,8 @@ fn test_function_calls() -> Result<()> {
         let source = debug.source();
         let name = names.name.map(|s| s.into_owned());
         hook_output.lock().unwrap().push((name, source.what));
-        Ok(())
-    });
+        Ok(VmState::Continue)
+    })?;
 
     lua.load(
         r#"
@@ -75,14 +74,20 @@ fn test_function_calls() -> Result<()> {
 
     let output = output.lock().unwrap();
     if cfg!(feature = "luajit") && lua.load("jit.version_num").eval::<i64>()? >= 20100 {
+        #[cfg(not(force_memory_limit))]
+        assert_eq!(*output, vec![(None, "main"), (Some("len".to_string()), "Lua")]);
+        #[cfg(force_memory_limit)]
         assert_eq!(
             *output,
-            vec![(None, "main"), (Some("len".to_string()), "Lua")]
+            vec![(None, "C"), (None, "main"), (Some("len".to_string()), "Lua")]
         );
     } else {
+        #[cfg(not(force_memory_limit))]
+        assert_eq!(*output, vec![(None, "main"), (Some("len".to_string()), "C")]);
+        #[cfg(force_memory_limit)]
         assert_eq!(
             *output,
-            vec![(None, "main"), (Some("len".to_string()), "C")]
+            vec![(None, "C"), (None, "main"), (Some("len".to_string()), "C")]
         );
     }
 
@@ -95,20 +100,13 @@ fn test_error_within_hook() -> Result<()> {
 
     lua.set_hook(HookTriggers::EVERY_LINE, |_lua, _debug| {
         Err(Error::runtime("Something happened in there!"))
-    });
-
-    let err = lua
-        .load("x = 1")
-        .exec()
-        .expect_err("panic didn't propagate");
+    })?;
 
+    let err = lua.load("x = 1").exec().expect_err("panic didn't propagate");
     match err {
-        Error::CallbackError { cause, .. } => match cause.deref() {
-            Error::RuntimeError(s) => assert_eq!(s, "Something happened in there!"),
-            _ => panic!("wrong callback error kind caught"),
-        },
-        _ => panic!("wrong error kind caught"),
-    };
+        Error::RuntimeError(msg) => assert_eq!(msg, "Something happened in there!"),
+        err => panic!("expected `RuntimeError` with a specific message, got {err:?}"),
+    }
 
     Ok(())
 }
@@ -129,10 +127,10 @@ fn test_limit_execution_instructions() -> Result<()> {
             if max_instructions.fetch_sub(30, Ordering::Relaxed) <= 30 {
                 Err(Error::runtime("time's up"))
             } else {
-                Ok(())
+                Ok(VmState::Continue)
             }
         },
-    );
+    )?;
 
     lua.globals().set("x", Value::Integer(0))?;
     let _ = lua
@@ -153,14 +151,9 @@ fn test_limit_execution_instructions() -> Result<()> {
 fn test_hook_removal() -> Result<()> {
     let lua = Lua::new();
 
-    lua.set_hook(
-        HookTriggers::new().every_nth_instruction(1),
-        |_lua, _debug| {
-            Err(Error::runtime(
-                "this hook should've been removed by this time",
-            ))
-        },
-    );
+    lua.set_hook(HookTriggers::new().every_nth_instruction(1), |_lua, _debug| {
+        Err(Error::runtime("this hook should've been removed by this time"))
+    })?;
 
     assert!(lua.load("local x = 1").exec().is_err());
     lua.remove_hook();
@@ -189,9 +182,10 @@ fn test_hook_swap_within_hook() -> Result<()> {
             .set_hook(HookTriggers::EVERY_LINE, move |lua, _debug| {
                 lua.globals().set("ok", 1i64)?;
                 TL_LUA.with(|tl| {
-                    tl.borrow().as_ref().unwrap().set_hook(
-                        HookTriggers::EVERY_LINE,
-                        move |lua, _debug| {
+                    tl.borrow()
+                        .as_ref()
+                        .unwrap()
+                        .set_hook(HookTriggers::EVERY_LINE, move |lua, _debug| {
                             lua.load(
                                 r#"
                                     if ok ~= nil then
@@ -204,13 +198,12 @@ fn test_hook_swap_within_hook() -> Result<()> {
                             TL_LUA.with(|tl| {
                                 tl.borrow().as_ref().unwrap().remove_hook();
                             });
-                            Ok(())
-                        },
-                    )
-                });
-                Ok(())
+                            Ok(VmState::Continue)
+                        })
+                })?;
+                Ok(VmState::Continue)
             })
-    });
+    })?;
 
     TL_LUA.with(|tl| {
         let tl = tl.borrow();
@@ -223,7 +216,7 @@ fn test_hook_swap_within_hook() -> Result<()> {
             "#,
         )
         .exec()?;
-        assert_eq!(lua.globals().get::<_, i64>("ok")?, 2);
+        assert_eq!(lua.globals().get::<i64>("ok")?, 2);
         Ok(())
     })
 }
@@ -247,11 +240,11 @@ fn test_hook_threads() -> Result<()> {
     let hook_output = output.clone();
     co.set_hook(HookTriggers::EVERY_LINE, move |_lua, debug| {
         assert_eq!(debug.event(), DebugEvent::Line);
-        hook_output.lock().unwrap().push(debug.curr_line());
-        Ok(())
-    });
+        hook_output.lock().unwrap().push(debug.current_line().unwrap());
+        Ok(VmState::Continue)
+    })?;
 
-    co.resume(())?;
+    co.resume::<()>(())?;
     lua.remove_hook();
 
     let output = output.lock().unwrap();
@@ -263,3 +256,72 @@ fn test_hook_threads() -> Result<()> {
 
     Ok(())
 }
+
+#[test]
+fn test_hook_yield() -> Result<()> {
+    let lua = Lua::new();
+
+    let func = lua
+        .load(
+            r#"
+            local x = 2 + 3
+            local y = x * 63
+            local z = string.len(x..", "..y)
+        "#,
+        )
+        .into_function()?;
+    let co = lua.create_thread(func)?;
+
+    co.set_hook(HookTriggers::EVERY_LINE, move |_lua, _debug| Ok(VmState::Yield))?;
+
+    #[cfg(any(feature = "lua54", feature = "lua53"))]
+    {
+        assert!(co.resume::<()>(()).is_ok());
+        assert!(co.resume::<()>(()).is_ok());
+        assert!(co.resume::<()>(()).is_ok());
+        assert!(co.resume::<()>(()).is_ok());
+        assert!(co.status() == ThreadStatus::Finished);
+    }
+    #[cfg(any(feature = "lua51", feature = "lua52", feature = "luajit"))]
+    {
+        assert!(
+            matches!(co.resume::<()>(()), Err(Error::RuntimeError(err)) if err.contains("attempt to yield from a hook"))
+        );
+        assert!(co.status() == ThreadStatus::Error);
+    }
+
+    Ok(())
+}
+
+#[test]
+fn test_global_hook() -> Result<()> {
+    let lua = Lua::new();
+
+    let counter = Arc::new(AtomicI64::new(0));
+    let hook_counter = counter.clone();
+    lua.set_global_hook(HookTriggers::EVERY_LINE, move |_lua, debug| {
+        assert_eq!(debug.event(), DebugEvent::Line);
+        hook_counter.fetch_add(1, Ordering::Relaxed);
+        Ok(VmState::Continue)
+    })?;
+
+    let thread = lua.create_thread(
+        lua.load(
+            r#"
+            local x = 2 + 3
+            local y = x * 63
+            coroutine.yield()
+            local z = string.len(x..", "..y)
+        "#,
+        )
+        .into_function()?,
+    )?;
+
+    thread.resume::<()>(()).unwrap();
+    lua.remove_global_hook();
+    thread.resume::<()>(()).unwrap();
+    assert_eq!(thread.status(), ThreadStatus::Finished);
+    assert_eq!(counter.load(Ordering::Relaxed), 3);
+
+    Ok(())
+}
