@@ -1,10 +1,11 @@
-#![cfg(feature = "serialize")]
+#![cfg(feature = "serde")]
 
 use std::collections::HashMap;
 use std::error::Error as StdError;
 
+use bstr::BString;
 use mlua::{
-    DeserializeOptions, Error, ExternalResult, Lua, LuaSerdeExt, Result as LuaResult,
+    AnyUserData, DeserializeOptions, Error, ExternalResult, IntoLua, Lua, LuaSerdeExt, Result as LuaResult,
     SerializeOptions, UserData, Value,
 };
 use serde::{Deserialize, Serialize};
@@ -24,7 +25,7 @@ fn test_serialize() -> Result<(), Box<dyn StdError>> {
     globals.set("null", lua.null())?;
 
     let empty_array = lua.create_table()?;
-    empty_array.set_metatable(Some(lua.array_metatable()));
+    empty_array.set_metatable(Some(lua.array_metatable()))?;
     globals.set("empty_array", empty_array)?;
 
     let val = lua
@@ -72,62 +73,30 @@ fn test_serialize() -> Result<(), Box<dyn StdError>> {
 }
 
 #[test]
-fn test_serialize_in_scope() -> LuaResult<()> {
-    #[derive(Serialize, Clone)]
-    struct MyUserData(i64, String);
-
-    impl UserData for MyUserData {}
-
-    let lua = Lua::new();
-    lua.scope(|scope| {
-        let ud = scope.create_ser_userdata(MyUserData(-5, "test userdata".into()))?;
-        assert_eq!(
-            serde_json::to_value(&ud).unwrap(),
-            serde_json::json!((-5, "test userdata"))
-        );
-        Ok(())
-    })?;
-
-    lua.scope(|scope| {
-        let ud = scope.create_ser_userdata(MyUserData(-5, "test userdata".into()))?;
-        lua.globals().set("ud", ud)
-    })?;
-    let val = lua.load("ud").eval::<Value>()?;
-    match serde_json::to_value(&val) {
-        Ok(v) => panic!("expected destructed error, got {}", v),
-        Err(e) if e.to_string().contains("destructed") => {}
-        Err(e) => panic!("expected destructed error, got {}", e),
-    }
-
-    struct MyUserDataRef<'a>(#[allow(unused)] &'a ());
-
-    impl<'a> UserData for MyUserDataRef<'a> {}
-
-    lua.scope(|scope| {
-        let ud = scope.create_nonstatic_userdata(MyUserDataRef(&()))?;
-        match serde_json::to_value(&ud) {
-            Ok(v) => panic!("expected serialization error, got {}", v),
-            Err(serde_json::Error { .. }) => {}
-        };
-        Ok(())
-    })?;
-
-    Ok(())
-}
-
-#[test]
-fn test_serialize_any_userdata() -> Result<(), Box<dyn StdError>> {
+fn test_serialize_any_userdata() {
     let lua = Lua::new();
 
     let json_val = serde_json::json!({
         "a": 1,
         "b": "test",
     });
-    let json_ud = lua.create_ser_any_userdata(json_val)?;
-    let json_str = serde_json::to_string_pretty(&json_ud)?;
+    let json_ud = lua.create_ser_any_userdata(json_val).unwrap();
+    let json_str = serde_json::to_string_pretty(&json_ud).unwrap();
     assert_eq!(json_str, "{\n  \"a\": 1,\n  \"b\": \"test\"\n}");
+}
 
-    Ok(())
+#[test]
+fn test_serialize_wrapped_any_userdata() {
+    let lua = Lua::new();
+
+    let json_val = serde_json::json!({
+        "a": 1,
+        "b": "test",
+    });
+    let ud = AnyUserData::wrap_ser(json_val);
+    let json_ud = ud.into_lua(&lua).unwrap();
+    let json_str = serde_json::to_string(&json_ud).unwrap();
+    assert_eq!(json_str, "{\"a\":1,\"b\":\"test\"}");
 }
 
 #[test]
@@ -165,13 +134,7 @@ fn test_serialize_failure() -> Result<(), Box<dyn StdError>> {
 fn test_serialize_vector() -> Result<(), Box<dyn StdError>> {
     let lua = Lua::new();
 
-    let globals = lua.globals();
-    globals.set(
-        "vector",
-        lua.create_function(|_, (x, y, z)| Ok(mlua::Vector::new(x, y, z)))?,
-    )?;
-
-    let val = lua.load("{_vector = vector(1, 2, 3)}").eval::<Value>()?;
+    let val = lua.load("{_vector = vector.create(1, 2, 3)}").eval::<Value>()?;
     let json = serde_json::json!({
         "_vector": [1.0, 2.0, 3.0],
     });
@@ -188,13 +151,9 @@ fn test_serialize_vector() -> Result<(), Box<dyn StdError>> {
 fn test_serialize_vector() -> Result<(), Box<dyn StdError>> {
     let lua = Lua::new();
 
-    let globals = lua.globals();
-    globals.set(
-        "vector",
-        lua.create_function(|_, (x, y, z, w)| Ok(mlua::Vector::new(x, y, z, w)))?,
-    )?;
-
-    let val = lua.load("{_vector = vector(1, 2, 3, 4)}").eval::<Value>()?;
+    let val = lua
+        .load("{_vector = vector.create(1, 2, 3, 4)}")
+        .eval::<Value>()?;
     let json = serde_json::json!({
         "_vector": [1.0, 2.0, 3.0, 4.0],
     });
@@ -214,7 +173,7 @@ fn test_serialize_sorted() -> LuaResult<()> {
     globals.set("null", lua.null())?;
 
     let empty_array = lua.create_table()?;
-    empty_array.set_metatable(Some(lua.array_metatable()));
+    empty_array.set_metatable(Some(lua.array_metatable()))?;
     globals.set("empty_array", empty_array)?;
 
     let value = lua
@@ -290,6 +249,65 @@ fn test_serialize_same_table_twice() -> LuaResult<()> {
     Ok(())
 }
 
+#[test]
+fn test_serialize_empty_table() -> LuaResult<()> {
+    let lua = Lua::new();
+
+    let table = Value::Table(lua.create_table()?);
+    let json = serde_json::to_string(&table.to_serializable()).unwrap();
+    assert_eq!(json, "{}");
+
+    // Set the option to encode empty tables as array
+    let json = serde_json::to_string(&table.to_serializable().encode_empty_tables_as_array(true)).unwrap();
+    assert_eq!(json, "[]");
+
+    // Check hashmap table with this option
+    table.as_table().unwrap().set("hello", "world")?;
+    let json = serde_json::to_string(&table.to_serializable().encode_empty_tables_as_array(true)).unwrap();
+    assert_eq!(json, r#"{"hello":"world"}"#);
+
+    Ok(())
+}
+
+#[test]
+fn test_serialize_mixed_table() -> LuaResult<()> {
+    let lua = Lua::new();
+
+    // Check that sparse array is serialized similarly when using direct serialization
+    // and via `Lua::from_value`
+    let table = lua.load("{1,2,3,nil,5}").eval::<Value>()?;
+    let json1 = serde_json::to_string(&table).unwrap();
+    let json2 = lua.from_value::<serde_json::Value>(table)?;
+    assert_eq!(json1, json2.to_string());
+
+    // A table with several borders should be correctly encoded when `detect_mixed_tables` is enabled
+    let table = lua
+        .load(
+            r#"
+        local t = {1,2,3,nil,5,6}
+        t[10] = 10
+        return t
+    "#,
+        )
+        .eval::<Value>()?;
+    let json = serde_json::to_string(&table.to_serializable().detect_mixed_tables(true)).unwrap();
+    assert_eq!(json, r#"[1,2,3,null,5,6,null,null,null,10]"#);
+
+    // A mixed table with both array-like and map-like entries
+    let table = lua.load(r#"{1,2,3, key="value"}"#).eval::<Value>()?;
+    let json = serde_json::to_string(&table).unwrap();
+    assert_eq!(json, r#"[1,2,3]"#);
+    let json = serde_json::to_string(&table.to_serializable().detect_mixed_tables(true)).unwrap();
+    assert_eq!(json, r#"{"1":1,"2":2,"3":3,"key":"value"}"#);
+
+    // A mixed table with duplicate keys of different types
+    let table = lua.load(r#"{1,2,3, ["1"]="value"}"#).eval::<Value>()?;
+    let json = serde_json::to_string(&table.to_serializable().detect_mixed_tables(true)).unwrap();
+    assert_eq!(json, r#"{"1":1,"2":2,"3":3,"1":"value"}"#);
+
+    Ok(())
+}
+
 #[test]
 fn test_to_value_struct() -> LuaResult<()> {
     let lua = Lua::new();
@@ -395,10 +413,7 @@ fn test_to_value_with_options() -> Result<(), Box<dyn StdError>> {
         unit: (),
         unitstruct: UnitStruct,
     };
-    let data2 = lua.to_value_with(
-        &mydata,
-        SerializeOptions::new().serialize_none_to_null(false),
-    )?;
+    let data2 = lua.to_value_with(&mydata, SerializeOptions::new().serialize_none_to_null(false))?;
     globals.set("data2", data2)?;
     lua.load(
         r#"
@@ -410,10 +425,7 @@ fn test_to_value_with_options() -> Result<(), Box<dyn StdError>> {
     .exec()?;
 
     // serialize_unit_to_null
-    let data3 = lua.to_value_with(
-        &mydata,
-        SerializeOptions::new().serialize_unit_to_null(false),
-    )?;
+    let data3 = lua.to_value_with(&mydata, SerializeOptions::new().serialize_unit_to_null(false))?;
     globals.set("data3", data3)?;
     lua.load(
         r#"
@@ -468,6 +480,7 @@ fn test_from_value_struct() -> Result<(), Box<dyn StdError>> {
         map: HashMap<i32, i32>,
         empty: Vec<()>,
         tuple: (u8, u8, u8),
+        bytes: BString,
     }
 
     let value = lua
@@ -479,6 +492,7 @@ fn test_from_value_struct() -> Result<(), Box<dyn StdError>> {
                 map = {2, [4] = 1},
                 empty = {},
                 tuple = {10, 20, 30},
+                bytes = "\240\040\140\040",
             }
         "#,
         )
@@ -491,6 +505,7 @@ fn test_from_value_struct() -> Result<(), Box<dyn StdError>> {
             map: vec![(1, 2), (4, 1)].into_iter().collect(),
             empty: vec![],
             tuple: (10, 20, 30),
+            bytes: BString::from([240, 40, 140, 40]),
         },
         got
     );
@@ -711,6 +726,37 @@ fn test_from_value_userdata() -> Result<(), Box<dyn StdError>> {
     Ok(())
 }
 
+#[test]
+fn test_from_value_empty_table() -> Result<(), Box<dyn StdError>> {
+    let lua = Lua::new();
+
+    // By default we encode empty tables as objects
+    let t = lua.create_table()?;
+    let got = lua.from_value::<serde_json::Value>(Value::Table(t.clone()))?;
+    assert_eq!(got, serde_json::json!({}));
+
+    // Set the option to encode empty tables as array
+    let got = lua
+        .from_value_with::<serde_json::Value>(
+            Value::Table(t.clone()),
+            DeserializeOptions::new().encode_empty_tables_as_array(true),
+        )
+        .unwrap();
+    assert_eq!(got, serde_json::json!([]));
+
+    // Check hashmap table with this option
+    t.raw_set("hello", "world")?;
+    let got = lua
+        .from_value_with::<serde_json::Value>(
+            Value::Table(t),
+            DeserializeOptions::new().encode_empty_tables_as_array(true),
+        )
+        .unwrap();
+    assert_eq!(got, serde_json::json!({"hello": "world"}));
+
+    Ok(())
+}
+
 #[test]
 fn test_from_value_sorted() -> Result<(), Box<dyn StdError>> {
     let lua = Lua::new();
@@ -767,27 +813,29 @@ fn test_arbitrary_precision() {
 
 #[cfg(feature = "luau")]
 #[test]
-fn test_buffer_serialize() {
+fn test_buffer_serialize() -> LuaResult<()> {
     let lua = Lua::new();
 
-    let buf = lua.create_buffer(&[1, 2, 3, 4]).unwrap();
+    let buf = lua.create_buffer(&[1, 2, 3, 4])?;
     let val = serde_value::to_value(&buf).unwrap();
     assert_eq!(val, serde_value::Value::Bytes(vec![1, 2, 3, 4]));
 
     // Try empty buffer
-    let buf = lua.create_buffer(&[]).unwrap();
+    let buf = lua.create_buffer(&[])?;
     let val = serde_value::to_value(&buf).unwrap();
     assert_eq!(val, serde_value::Value::Bytes(vec![]));
+
+    Ok(())
 }
 
 #[cfg(feature = "luau")]
 #[test]
-fn test_buffer_from_value() {
+fn test_buffer_from_value() -> LuaResult<()> {
     let lua = Lua::new();
 
-    let buf = lua.create_buffer(&[1, 2, 3, 4]).unwrap();
-    let val = lua
-        .from_value::<serde_value::Value>(Value::UserData(buf))
-        .unwrap();
+    let buf = lua.create_buffer(&[1, 2, 3, 4])?;
+    let val = lua.from_value::<serde_value::Value>(Value::Buffer(buf)).unwrap();
     assert_eq!(val, serde_value::Value::Bytes(vec![1, 2, 3, 4]));
+
+    Ok(())
 }
