@@ -3,21 +3,38 @@ use std::collections::HashMap;
 use std::iter::FromIterator;
 use std::panic::{catch_unwind, AssertUnwindSafe};
 use std::string::String as StdString;
-use std::sync::atomic::{AtomicU32, Ordering};
 use std::sync::Arc;
 use std::{error, f32, f64, fmt};
 
 use mlua::{
-    ChunkMode, Error, ExternalError, Function, Lua, LuaOptions, Nil, Result, StdLib, String, Table,
+    ffi, ChunkMode, Error, ExternalError, Function, Lua, LuaOptions, Nil, Result, StdLib, String, Table,
     UserData, Value, Variadic,
 };
 
+#[test]
+fn test_weak_lua() {
+    let lua = Lua::new();
+    let weak_lua = lua.weak();
+    assert!(weak_lua.try_upgrade().is_some());
+    drop(lua);
+    assert!(weak_lua.try_upgrade().is_none());
+}
+
+#[test]
+#[should_panic(expected = "Lua instance is destroyed")]
+fn test_weak_lua_panic() {
+    let lua = Lua::new();
+    let weak_lua = lua.weak();
+    drop(lua);
+    let _ = weak_lua.upgrade();
+}
+
 #[cfg(not(feature = "luau"))]
 #[test]
 fn test_safety() -> Result<()> {
     let lua = Lua::new();
     assert!(lua.load(r#"require "debug""#).exec().is_err());
-    match lua.load_from_std_lib(StdLib::DEBUG) {
+    match lua.load_std_libs(StdLib::DEBUG) {
         Err(Error::SafetyError(_)) => {}
         Err(e) => panic!("expected SafetyError, got {:?}", e),
         Ok(_) => panic!("expected SafetyError, got no error"),
@@ -52,8 +69,8 @@ fn test_safety() -> Result<()> {
 
     // Test safety rules after dynamically loading `package` library
     let lua = Lua::new_with(StdLib::NONE, LuaOptions::default())?;
-    assert!(lua.globals().get::<_, Option<Value>>("require")?.is_none());
-    lua.load_from_std_lib(StdLib::PACKAGE)?;
+    assert!(lua.globals().get::<Option<Value>>("require")?.is_none());
+    lua.load_std_libs(StdLib::PACKAGE)?;
     match lua.load(r#"package.loadlib()"#).exec() {
         Err(Error::CallbackError { ref cause, .. }) => match cause.as_ref() {
             Error::SafetyError(_) => {}
@@ -91,7 +108,7 @@ fn test_exec() -> Result<()> {
     "#,
     )
     .exec()?;
-    assert_eq!(globals.get::<_, String>("res")?, "foobar");
+    assert_eq!(globals.get::<String>("res")?, "foobar");
 
     let module: Table = lua
         .load(
@@ -106,12 +123,8 @@ fn test_exec() -> Result<()> {
         "#,
         )
         .eval()?;
-    println!("checkpoint");
     assert!(module.contains_key("func")?);
-    assert_eq!(
-        module.get::<_, Function>("func")?.call::<_, String>(())?,
-        "hello"
-    );
+    assert_eq!(module.get::<Function>("func")?.call::<String>(())?, "hello");
 
     Ok(())
 }
@@ -128,10 +141,32 @@ fn test_eval() -> Result<()> {
             incomplete_input: true,
             ..
         }) => {}
-        r => panic!(
-            "expected SyntaxError with incomplete_input=true, got {:?}",
-            r
-        ),
+        r => panic!("expected SyntaxError with incomplete_input=true, got {:?}", r),
+    }
+
+    Ok(())
+}
+
+#[test]
+fn test_replace_globals() -> Result<()> {
+    let lua = Lua::new();
+
+    let globals = lua.create_table()?;
+    globals.set("foo", "bar")?;
+
+    lua.set_globals(globals.clone())?;
+    let val = lua.load("return foo").eval::<StdString>()?;
+    assert_eq!(val, "bar");
+
+    // Updating globals in sandboxed Lua state is not allowed
+    #[cfg(feature = "luau")]
+    {
+        lua.sandbox(true)?;
+        match lua.set_globals(globals) {
+            Err(Error::RuntimeError(msg))
+                if msg.contains("cannot change globals in a sandboxed Lua state") => {}
+            r => panic!("expected RuntimeError(...) with a specific error message, got {r:?}"),
+        }
     }
 
     Ok(())
@@ -141,10 +176,7 @@ fn test_eval() -> Result<()> {
 fn test_load_mode() -> Result<()> {
     let lua = unsafe { Lua::unsafe_new() };
 
-    assert_eq!(
-        lua.load("1 + 1").set_mode(ChunkMode::Text).eval::<i32>()?,
-        2
-    );
+    assert_eq!(lua.load("1 + 1").set_mode(ChunkMode::Text).eval::<i32>()?, 2);
     match lua.load("1 + 1").set_mode(ChunkMode::Binary).exec() {
         Ok(_) => panic!("expected SyntaxError, got no error"),
         Err(Error::SyntaxError { message: msg, .. }) => {
@@ -156,14 +188,9 @@ fn test_load_mode() -> Result<()> {
     #[cfg(not(feature = "luau"))]
     let bytecode = lua.load("return 1 + 1").into_function()?.dump(true);
     #[cfg(feature = "luau")]
-    let bytecode = mlua::Compiler::new().compile("return 1 + 1");
+    let bytecode = mlua::Compiler::new().compile("return 1 + 1")?;
     assert_eq!(lua.load(&bytecode).eval::<i32>()?, 2);
-    assert_eq!(
-        lua.load(&bytecode)
-            .set_mode(ChunkMode::Binary)
-            .eval::<i32>()?,
-        2
-    );
+    assert_eq!(lua.load(&bytecode).set_mode(ChunkMode::Binary).eval::<i32>()?, 2);
     match lua.load(&bytecode).set_mode(ChunkMode::Text).exec() {
         Ok(_) => panic!("expected SyntaxError, got no error"),
         Err(Error::SyntaxError { message: msg, .. }) => {
@@ -193,13 +220,13 @@ fn test_lua_multi() -> Result<()> {
     .exec()?;
 
     let globals = lua.globals();
-    let concat = globals.get::<_, Function>("concat")?;
-    let mreturn = globals.get::<_, Function>("mreturn")?;
+    let concat = globals.get::<Function>("concat")?;
+    let mreturn = globals.get::<Function>("mreturn")?;
 
-    assert_eq!(concat.call::<_, String>(("foo", "bar"))?, "foobar");
-    let (a, b) = mreturn.call::<_, (u64, u64)>(())?;
+    assert_eq!(concat.call::<String>(("foo", "bar"))?, "foobar");
+    let (a, b) = mreturn.call::<(u64, u64)>(())?;
     assert_eq!((a, b), (1, 2));
-    let (a, b, v) = mreturn.call::<_, (u64, u64, Variadic<u64>)>(())?;
+    let (a, b, v) = mreturn.call::<(u64, u64, Variadic<u64>)>(())?;
     assert_eq!((a, b), (1, 2));
     assert_eq!(v[..], [3, 4, 5, 6]);
 
@@ -221,10 +248,10 @@ fn test_coercion() -> Result<()> {
     .exec()?;
 
     let globals = lua.globals();
-    assert_eq!(globals.get::<_, String>("int")?, "123");
-    assert_eq!(globals.get::<_, i32>("str")?, 123);
-    assert_eq!(globals.get::<_, i32>("num")?, 123);
-    assert!(globals.get::<_, String>("func").is_err());
+    assert_eq!(globals.get::<String>("int")?, "123");
+    assert_eq!(globals.get::<i32>("str")?, 123);
+    assert_eq!(globals.get::<i32>("num")?, 123);
+    assert!(globals.get::<String>("func").is_err());
 
     Ok(())
 }
@@ -308,40 +335,36 @@ fn test_error() -> Result<()> {
     )
     .exec()?;
 
-    let rust_error_function =
-        lua.create_function(|_, ()| -> Result<()> { Err(TestError.into_lua_err()) })?;
+    let rust_error_function = lua.create_function(|_, ()| -> Result<()> { Err(TestError.into_lua_err()) })?;
     globals.set("rust_error_function", rust_error_function)?;
 
-    let no_error = globals.get::<_, Function>("no_error")?;
-    assert!(no_error.call::<_, ()>(()).is_ok());
+    let no_error = globals.get::<Function>("no_error")?;
+    assert!(no_error.call::<()>(()).is_ok());
 
-    let lua_error = globals.get::<_, Function>("lua_error")?;
-    match lua_error.call::<_, ()>(()) {
+    let lua_error = globals.get::<Function>("lua_error")?;
+    match lua_error.call::<()>(()) {
         Err(Error::RuntimeError(_)) => {}
         Err(e) => panic!("error is not RuntimeError kind, got {:?}", e),
         _ => panic!("error not returned"),
     }
 
-    let rust_error = globals.get::<_, Function>("rust_error")?;
-    match rust_error.call::<_, ()>(()) {
+    let rust_error = globals.get::<Function>("rust_error")?;
+    match rust_error.call::<()>(()) {
         Err(Error::CallbackError { .. }) => {}
         Err(e) => panic!("error is not CallbackError kind, got {:?}", e),
         _ => panic!("error not returned"),
     }
 
-    let return_error = globals.get::<_, Function>("return_error")?;
-    match return_error.call::<_, Value>(()) {
+    let return_error = globals.get::<Function>("return_error")?;
+    match return_error.call::<Value>(()) {
         Ok(Value::Error(_)) => {}
         _ => panic!("Value::Error not returned"),
     }
 
-    let return_string_error = globals.get::<_, Function>("return_string_error")?;
-    assert!(return_string_error.call::<_, Error>(()).is_ok());
+    let return_string_error = globals.get::<Function>("return_string_error")?;
+    assert!(return_string_error.call::<Error>(()).is_ok());
 
-    match lua
-        .load("if youre happy and you know it syntax error")
-        .exec()
-    {
+    match lua.load("if you are happy and you know it syntax error").exec() {
         Err(Error::SyntaxError {
             incomplete_input: false,
             ..
@@ -358,31 +381,30 @@ fn test_error() -> Result<()> {
         _ => panic!("error not returned"),
     }
 
-    let test_pcall = globals.get::<_, Function>("test_pcall")?;
-    test_pcall.call::<_, ()>(())?;
+    let test_pcall = globals.get::<Function>("test_pcall")?;
+    test_pcall.call::<()>(())?;
 
     #[cfg(not(target_arch = "wasm32"))]
     {
-        let understand_recursion = globals.get::<_, Function>("understand_recursion")?;
-        assert!(understand_recursion.call::<_, ()>(()).is_err());
+        let understand_recursion = globals.get::<Function>("understand_recursion")?;
+        assert!(understand_recursion.call::<()>(()).is_err());
     }
 
     Ok(())
 }
 
 #[test]
+#[cfg(not(panic = "abort"))]
 fn test_panic() -> Result<()> {
     fn make_lua(options: LuaOptions) -> Result<Lua> {
         let lua = Lua::new_with(StdLib::ALL_SAFE, options)?;
-        let rust_panic_function =
-            lua.create_function(|_, msg: Option<StdString>| -> Result<()> {
-                if let Some(msg) = msg {
-                    panic!("{}", msg)
-                }
-                panic!("rust panic")
-            })?;
-        lua.globals()
-            .set("rust_panic_function", rust_panic_function)?;
+        let rust_panic_function = lua.create_function(|_, msg: Option<StdString>| -> Result<()> {
+            if let Some(msg) = msg {
+                panic!("{}", msg)
+            }
+            panic!("rust panic")
+        })?;
+        lua.globals().set("rust_panic_function", rust_panic_function)?;
         Ok(lua)
     }
 
@@ -416,7 +438,7 @@ fn test_panic() -> Result<()> {
     {
         let lua = make_lua(LuaOptions::default())?;
         match catch_unwind(AssertUnwindSafe(|| -> Result<()> {
-            let _catched_panic = lua
+            let _caught_panic = lua
                 .load(
                     r#"
                     -- Set global
@@ -431,7 +453,7 @@ fn test_panic() -> Result<()> {
             Err(_) => {}
         };
 
-        assert!(lua.globals().get::<_, Value>("err")? == Value::Nil);
+        assert!(lua.globals().get::<Value>("err")? == Value::Nil);
         match lua.load("tostring(err)").exec() {
             Ok(_) => panic!("no error was detected"),
             Err(Error::CallbackError { ref cause, .. }) => match cause.as_ref() {
@@ -504,6 +526,29 @@ fn test_panic() -> Result<()> {
     Ok(())
 }
 
+#[cfg(target_pointer_width = "64")]
+#[test]
+fn test_safe_integers() -> Result<()> {
+    const MAX_SAFE_INTEGER: i64 = 2i64.pow(53) - 1;
+    const MIN_SAFE_INTEGER: i64 = -2i64.pow(53) + 1;
+
+    let lua = Lua::new();
+    let f = lua.load("return ...").into_function()?;
+
+    assert_eq!(f.call::<i64>(MAX_SAFE_INTEGER)?, MAX_SAFE_INTEGER);
+    assert_eq!(f.call::<i64>(MIN_SAFE_INTEGER)?, MIN_SAFE_INTEGER);
+
+    // For Lua versions that does not support 64-bit integers, the values will be converted to f64
+    #[cfg(any(feature = "luau", feature = "lua51", feature = "luajit"))]
+    {
+        assert_ne!(f.call::<i64>(MAX_SAFE_INTEGER + 2)?, MAX_SAFE_INTEGER + 2);
+        assert_ne!(f.call::<i64>(MIN_SAFE_INTEGER - 2)?, MIN_SAFE_INTEGER - 2);
+        assert_eq!(f.call::<f64>(i64::MAX)?, i64::MAX as f64);
+    }
+
+    Ok(())
+}
+
 #[test]
 fn test_num_conversion() -> Result<()> {
     let lua = Lua::new();
@@ -538,12 +583,7 @@ fn test_num_conversion() -> Result<()> {
     assert_eq!(lua.load("1.0").eval::<f64>()?, 1.0);
     #[cfg(any(feature = "lua54", feature = "lua53"))]
     assert_eq!(lua.load("1.0").eval::<String>()?, "1.0");
-    #[cfg(any(
-        feature = "lua52",
-        feature = "lua51",
-        feature = "luajit",
-        feature = "luau"
-    ))]
+    #[cfg(any(feature = "lua52", feature = "lua51", feature = "luajit", feature = "luau"))]
     assert_eq!(lua.load("1.0").eval::<String>()?, "1");
 
     assert_eq!(lua.load("1.5").eval::<i64>()?, 1);
@@ -563,6 +603,21 @@ fn test_num_conversion() -> Result<()> {
 
     assert_eq!(lua.unpack::<i128>(lua.pack(1i128 << 64)?)?, 1i128 << 64);
 
+    // Negative zero
+    let negative_zero = lua.load("-0.0").eval::<f64>()?;
+    assert_eq!(negative_zero, 0.0);
+    // LuaJIT treats -0.0 as a positive zero
+    #[cfg(not(feature = "luajit"))]
+    assert!(negative_zero.is_sign_negative());
+
+    // In Lua <5.3 all numbers are floats
+    #[cfg(not(any(feature = "lua54", feature = "lua53", feature = "luajit")))]
+    {
+        let negative_zero = lua.load("-0").eval::<f64>()?;
+        assert_eq!(negative_zero, 0.0);
+        assert!(negative_zero.is_sign_negative());
+    }
+
     Ok(())
 }
 
@@ -616,23 +671,15 @@ fn test_pcall_xpcall() -> Result<()> {
     )
     .exec()?;
 
-    assert_eq!(globals.get::<_, bool>("pcall_status")?, false);
-    assert_eq!(globals.get::<_, String>("pcall_error")?, "testerror");
+    assert_eq!(globals.get::<bool>("pcall_status")?, false);
+    assert_eq!(globals.get::<String>("pcall_error")?, "testerror");
 
-    assert_eq!(globals.get::<_, bool>("xpcall_statusr")?, false);
-    #[cfg(any(
-        feature = "lua54",
-        feature = "lua53",
-        feature = "lua52",
-        feature = "luajit"
-    ))]
-    assert_eq!(
-        globals.get::<_, std::string::String>("xpcall_error")?,
-        "testerror"
-    );
+    assert_eq!(globals.get::<bool>("xpcall_statusr")?, false);
+    #[cfg(any(feature = "lua54", feature = "lua53", feature = "lua52", feature = "luajit"))]
+    assert_eq!(globals.get::<std::string::String>("xpcall_error")?, "testerror");
     #[cfg(feature = "lua51")]
     assert!(globals
-        .get::<_, String>("xpcall_error")?
+        .get::<String>("xpcall_error")?
         .to_str()?
         .ends_with(": testerror"));
 
@@ -645,9 +692,7 @@ fn test_pcall_xpcall() -> Result<()> {
     "#,
     )
     .exec()?;
-    let _ = globals
-        .get::<_, Function>("xpcall_recursion")?
-        .call::<_, ()>(());
+    let _ = globals.get::<Function>("xpcall_recursion")?.call::<()>(());
 
     Ok(())
 }
@@ -657,23 +702,22 @@ fn test_recursive_mut_callback_error() -> Result<()> {
     let lua = Lua::new();
 
     let mut v = Some(Box::new(123));
-    let f = lua.create_function_mut::<_, (), _>(move |lua, mutate: bool| {
+    let f = lua.create_function_mut(move |lua, mutate: bool| {
         if mutate {
             v = None;
         } else {
             // Produce a mutable reference
             let r = v.as_mut().unwrap();
             // Whoops, this will recurse into the function and produce another mutable reference!
-            lua.globals().get::<_, Function>("f")?.call::<_, ()>(true)?;
+            lua.globals().get::<Function>("f")?.call::<()>(true)?;
             println!("Should not get here, mutable aliasing has occurred!");
-            println!("value at {:p}", r as *mut _);
-            println!("value is {}", r);
+            println!("value at {:p} is {r}", r as *mut _);
         }
 
         Ok(())
     })?;
     lua.globals().set("f", f)?;
-    match lua.globals().get::<_, Function>("f")?.call::<_, ()>(false) {
+    match lua.globals().get::<Function>("f")?.call::<()>(false) {
         Err(Error::CallbackError { ref cause, .. }) => match *cause.as_ref() {
             Error::CallbackError { ref cause, .. } => match *cause.as_ref() {
                 Error::RecursiveMutCallback { .. } => {}
@@ -704,13 +748,13 @@ fn test_set_metatable_nil() -> Result<()> {
 fn test_named_registry_value() -> Result<()> {
     let lua = Lua::new();
 
-    lua.set_named_registry_value::<i32>("test", 42)?;
+    lua.set_named_registry_value("test", 42)?;
     let f = lua.create_function(move |lua, ()| {
         assert_eq!(lua.named_registry_value::<i32>("test")?, 42);
         Ok(())
     })?;
 
-    f.call::<_, ()>(())?;
+    f.call::<()>(())?;
 
     lua.unset_named_registry_value("test")?;
     match lua.named_registry_value("test")? {
@@ -725,7 +769,7 @@ fn test_named_registry_value() -> Result<()> {
 fn test_registry_value() -> Result<()> {
     let lua = Lua::new();
 
-    let mut r = Some(lua.create_registry_value::<i32>(42)?);
+    let mut r = Some(lua.create_registry_value(42)?);
     let f = lua.create_function_mut(move |lua, ()| {
         if let Some(r) = r.take() {
             assert_eq!(lua.registry_value::<i32>(&r)?, 42);
@@ -736,7 +780,7 @@ fn test_registry_value() -> Result<()> {
         Ok(())
     })?;
 
-    f.call::<_, ()>(())?;
+    f.call::<()>(())?;
 
     Ok(())
 }
@@ -767,20 +811,19 @@ fn test_drop_registry_value() -> Result<()> {
 fn test_replace_registry_value() -> Result<()> {
     let lua = Lua::new();
 
-    let key = lua.create_registry_value::<i32>(42)?;
-    lua.replace_registry_value(&key, "new value")?;
+    let mut key = lua.create_registry_value(42)?;
+    lua.replace_registry_value(&mut key, "new value")?;
     assert_eq!(lua.registry_value::<String>(&key)?, "new value");
-    lua.replace_registry_value(&key, Value::Nil)?;
+    lua.replace_registry_value(&mut key, Value::Nil)?;
     assert_eq!(lua.registry_value::<Value>(&key)?, Value::Nil);
-    lua.replace_registry_value(&key, 123)?;
+    lua.replace_registry_value(&mut key, 123)?;
     assert_eq!(lua.registry_value::<i32>(&key)?, 123);
 
-    // It should be impossible to replace (initial) nil value with non-nil
-    let key2 = lua.create_registry_value(Value::Nil)?;
-    match lua.replace_registry_value(&key2, "abc") {
-        Err(Error::RuntimeError(_)) => {}
-        r => panic!("expected RuntimeError, got {r:?}"),
-    }
+    let mut key2 = lua.create_registry_value(Value::Nil)?;
+    lua.replace_registry_value(&mut key2, Value::Nil)?;
+    assert_eq!(lua.registry_value::<Value>(&key2)?, Value::Nil);
+    lua.replace_registry_value(&mut key2, "abc")?;
+    assert_eq!(lua.registry_value::<String>(&key2)?, "abc");
 
     Ok(())
 }
@@ -855,6 +898,7 @@ fn test_registry_value_reuse() -> Result<()> {
 }
 
 #[test]
+#[cfg(not(panic = "abort"))]
 fn test_application_data() -> Result<()> {
     let lua = Lua::new();
 
@@ -882,10 +926,12 @@ fn test_application_data() -> Result<()> {
     assert_eq!(format!("{s:?}"), "\"test1\"");
 
     // Borrowing immutably and mutably of the same type is not allowed
+    assert!(lua.try_app_data_mut::<&str>().is_err());
     match catch_unwind(AssertUnwindSafe(|| lua.app_data_mut::<&str>().unwrap())) {
         Ok(_) => panic!("expected panic"),
         Err(_) => {}
     }
+    assert!(lua.try_app_data_ref::<Vec<&str>>().is_err());
     drop((s, v));
 
     // Test that application data is accessible from anywhere
@@ -899,13 +945,10 @@ fn test_application_data() -> Result<()> {
 
         Ok(())
     })?;
-    f.call(())?;
+    f.call::<()>(())?;
 
     assert_eq!(*lua.app_data_ref::<&str>().unwrap(), "test4");
-    assert_eq!(
-        *lua.app_data_ref::<Vec<&str>>().unwrap(),
-        vec!["test2", "test3"]
-    );
+    assert_eq!(*lua.app_data_ref::<Vec<&str>>().unwrap(), vec!["test2", "test3"]);
 
     lua.remove_app_data::<Vec<&str>>();
     assert!(matches!(lua.app_data_ref::<Vec<&str>>(), None));
@@ -913,6 +956,51 @@ fn test_application_data() -> Result<()> {
     Ok(())
 }
 
+#[test]
+fn test_rust_function() -> Result<()> {
+    let lua = Lua::new();
+
+    let globals = lua.globals();
+    lua.load(
+        r#"
+        function lua_function()
+            return rust_function()
+        end
+
+        -- Test to make sure chunk return is ignored
+        return 1
+    "#,
+    )
+    .exec()?;
+
+    let lua_function = globals.get::<Function>("lua_function")?;
+    let rust_function = lua.create_function(|_, ()| Ok("hello"))?;
+
+    globals.set("rust_function", rust_function)?;
+    assert_eq!(lua_function.call::<String>(())?, "hello");
+
+    Ok(())
+}
+
+#[test]
+fn test_c_function() -> Result<()> {
+    let lua = Lua::new();
+
+    extern "C-unwind" fn c_function(state: *mut mlua::lua_State) -> std::os::raw::c_int {
+        unsafe {
+            ffi::lua_pushboolean(state, 1);
+            ffi::lua_setglobal(state, b"c_function\0" as *const _ as *const _);
+        }
+        0
+    }
+
+    let func = unsafe { lua.create_c_function(c_function)? };
+    func.call::<()>(())?;
+    assert_eq!(lua.globals().get::<bool>("c_function")?, true);
+
+    Ok(())
+}
+
 #[test]
 #[cfg(not(target_arch = "wasm32"))]
 fn test_recursion() -> Result<()> {
@@ -920,15 +1008,13 @@ fn test_recursion() -> Result<()> {
 
     let f = lua.create_function(move |lua, i: i32| {
         if i < 64 {
-            lua.globals()
-                .get::<_, Function>("f")?
-                .call::<_, ()>(i + 1)?;
+            lua.globals().get::<Function>("f")?.call::<()>(i + 1)?;
         }
         Ok(())
     })?;
 
     lua.globals().set("f", &f)?;
-    f.call::<_, ()>(1)?;
+    f.call::<()>(1)?;
 
     Ok(())
 }
@@ -938,7 +1024,7 @@ fn test_recursion() -> Result<()> {
 fn test_too_many_returns() -> Result<()> {
     let lua = Lua::new();
     let f = lua.create_function(|_, ()| Ok(Variadic::from_iter(1..1000000)))?;
-    assert!(f.call::<_, Variadic<u32>>(()).is_err());
+    assert!(f.call::<Variadic<u32>>(()).is_err());
     Ok(())
 }
 
@@ -948,11 +1034,7 @@ fn test_too_many_arguments() -> Result<()> {
     let lua = Lua::new();
     lua.load("function test(...) end").exec()?;
     let args = Variadic::from_iter(1..1000000);
-    assert!(lua
-        .globals()
-        .get::<_, Function>("test")?
-        .call::<_, ()>(args)
-        .is_err());
+    assert!(lua.globals().get::<Function>("test")?.call::<()>(args).is_err());
     Ok(())
 }
 
@@ -962,33 +1044,10 @@ fn test_too_many_arguments() -> Result<()> {
 fn test_too_many_recursions() -> Result<()> {
     let lua = Lua::new();
 
-    let f = lua
-        .create_function(move |lua, ()| lua.globals().get::<_, Function>("f")?.call::<_, ()>(()))?;
+    let f = lua.create_function(move |lua, ()| lua.globals().get::<Function>("f")?.call::<()>(()))?;
 
     lua.globals().set("f", &f)?;
-    assert!(f.call::<_, ()>(()).is_err());
-
-    Ok(())
-}
-
-#[test]
-#[cfg(not(target_arch = "wasm32"))]
-fn test_too_many_binds() -> Result<()> {
-    let lua = Lua::new();
-    let globals = lua.globals();
-    lua.load(
-        r#"
-        function f(...)
-        end
-    "#,
-    )
-    .exec()?;
-
-    let concat = globals.get::<_, Function>("f")?;
-    assert!(concat.bind(Variadic::from_iter(1..1000000)).is_err());
-    assert!(concat
-        .call::<_, ()>(Variadic::from_iter(1..1000000))
-        .is_err());
+    assert!(f.call::<()>(()).is_err());
 
     Ok(())
 }
@@ -1039,10 +1098,7 @@ fn test_large_args() -> Result<()> {
         )
         .eval()?;
 
-    assert_eq!(
-        f.call::<_, usize>((0..100).collect::<Variadic<usize>>())?,
-        4950
-    );
+    assert_eq!(f.call::<usize>((0..100).collect::<Variadic<usize>>())?, 4950);
 
     Ok(())
 }
@@ -1058,7 +1114,7 @@ fn test_large_args_ref() -> Result<()> {
         Ok(())
     })?;
 
-    f.call::<_, ()>((0..100).map(|i| i.to_string()).collect::<Variadic<_>>())?;
+    f.call::<()>((0..100).map(|i| i.to_string()).collect::<Variadic<_>>())?;
 
     Ok(())
 }
@@ -1111,20 +1167,15 @@ fn test_context_thread() -> Result<()> {
         )
         .into_function()?;
 
-    #[cfg(any(
-        feature = "lua54",
-        feature = "lua53",
-        feature = "lua52",
-        feature = "luajit52"
-    ))]
-    f.call::<_, ()>(lua.current_thread())?;
+    #[cfg(any(feature = "lua54", feature = "lua53", feature = "lua52", feature = "luajit52"))]
+    f.call::<()>(lua.current_thread())?;
 
     #[cfg(any(
         feature = "lua51",
         all(feature = "luajit", not(feature = "luajit52")),
         feature = "luau"
     ))]
-    f.call::<_, ()>(Nil)?;
+    f.call::<()>(Nil)?;
 
     Ok(())
 }
@@ -1145,7 +1196,7 @@ fn test_context_thread_51() -> Result<()> {
         .eval()?,
     )?;
 
-    thread.resume::<_, ()>(thread.clone())?;
+    thread.resume::<()>(thread.clone())?;
 
     Ok(())
 }
@@ -1155,44 +1206,95 @@ fn test_context_thread_51() -> Result<()> {
 fn test_jit_version() -> Result<()> {
     let lua = Lua::new();
     let jit: Table = lua.globals().get("jit")?;
-    assert!(jit
-        .get::<_, String>("version")?
-        .to_str()?
-        .contains("LuaJIT"));
+    assert!(jit.get::<String>("version")?.to_str()?.contains("LuaJIT"));
     Ok(())
 }
 
 #[test]
-fn test_load_from_function() -> Result<()> {
+fn test_register_module() -> Result<()> {
     let lua = Lua::new();
 
-    let i = Arc::new(AtomicU32::new(0));
-    let i2 = i.clone();
-    let func = lua.create_function(move |lua, modname: String| {
-        i2.fetch_add(1, Ordering::Relaxed);
+    let t = lua.create_table()?;
+    t.set("name", "my_module")?;
+    lua.register_module("@my_module", &t)?;
+
+    lua.load(
+        r#"
+        local my_module = require("@my_module")
+        assert(my_module.name == "my_module")
+    "#,
+    )
+    .exec()?;
+
+    lua.unload_module("@my_module")?;
+    lua.load(
+        r#"
+        local ok, err = pcall(function() return require("@my_module") end)
+        assert(not ok)
+        "#,
+    )
+    .exec()?;
+
+    #[cfg(feature = "luau")]
+    {
+        // Luau registered modules must have '@' prefix
+        let res = lua.register_module("my_module", 123);
+        assert!(res.is_err());
+        assert_eq!(
+            res.unwrap_err().to_string(),
+            "runtime error: module name must begin with '@'"
+        );
+
+        // Luau registered modules (aliases) are case-insensitive
+        let res = lua.register_module("@My_Module", &t);
+        assert!(res.is_ok());
+        lua.load(
+            r#"
+            local my_module = require("@MY_MODule")
+            assert(my_module.name == "my_module")
+        "#,
+        )
+        .exec()?;
+    }
+
+    Ok(())
+}
+
+#[test]
+#[cfg(not(feature = "luau"))]
+fn test_preload_module() -> Result<()> {
+    let lua = Lua::new();
+
+    let loader = lua.create_function(move |lua, modname: String| {
         let t = lua.create_table()?;
-        t.set("__name", modname)?;
+        t.set("name", modname)?;
         Ok(t)
     })?;
 
-    let t: Table = lua.load_from_function("my_module", func.clone())?;
-    assert_eq!(t.get::<_, String>("__name")?, "my_module");
-    assert_eq!(i.load(Ordering::Relaxed), 1);
-
-    let _: Value = lua.load_from_function("my_module", func.clone())?;
-    assert_eq!(i.load(Ordering::Relaxed), 1);
-
-    let func_nil = lua.create_function(move |_, _: String| Ok(Value::Nil))?;
-    let v: Value = lua.load_from_function("my_module2", func_nil)?;
-    assert_eq!(v, Value::Boolean(true));
+    lua.preload_module("@my_module", loader.clone())?;
+    lua.load(
+        r#"
+        -- `my_module` is global for purposes of next test
+        my_module = require("@my_module")
+        assert(my_module.name == "@my_module")
+        local my_module2 = require("@my_module")
+        assert(my_module == my_module2)
+    "#,
+    )
+    .exec()
+    .unwrap();
 
     // Test unloading and loading again
-    lua.unload("my_module")?;
-    let _: Value = lua.load_from_function("my_module", func)?;
-    assert_eq!(i.load(Ordering::Relaxed), 2);
-
-    // Unloading nonexistent module must not fail
-    lua.unload("my_module2")?;
+    lua.unload_module("@my_module")?;
+    lua.load(
+        r#"
+        local my_module3 = require("@my_module")
+        -- `my_module` is not equal to `my_module3` because it was reloaded
+        assert(my_module ~= my_module3)
+        "#,
+    )
+    .exec()
+    .unwrap();
 
     Ok(())
 }
@@ -1202,14 +1304,18 @@ fn test_inspect_stack() -> Result<()> {
     let lua = Lua::new();
 
     // Not inside any function
-    assert!(lua.inspect_stack(0).is_none());
+    assert!(lua.inspect_stack(0, |_| ()).is_none());
 
     let logline = lua.create_function(|lua, msg: StdString| {
-        let debug = lua.inspect_stack(1).unwrap(); // caller
-        let source = debug.source().short_src;
-        let source = source.as_deref().unwrap_or("?");
-        let line = debug.curr_line();
-        Ok(format!("{}:{} {}", source, line, msg))
+        let r = lua
+            .inspect_stack(1, |debug| {
+                let source = debug.source().short_src;
+                let source = source.as_deref().unwrap_or("?");
+                let line = debug.current_line().unwrap();
+                format!("{}:{} {}", source, line, msg)
+            })
+            .unwrap();
+        Ok(r)
     })?;
     lua.globals().set("logline", logline)?;
 
@@ -1231,6 +1337,130 @@ fn test_inspect_stack() -> Result<()> {
     .set_name("chunk")
     .exec()?;
 
+    let stack_info = lua.create_function(|lua, ()| {
+        let stack_info = lua.inspect_stack(1, |debug| debug.stack()).unwrap();
+        Ok(format!("{stack_info:?}"))
+    })?;
+    lua.globals().set("stack_info", stack_info)?;
+
+    #[cfg(any(feature = "lua54", feature = "lua53", feature = "lua52", feature = "luau"))]
+    lua.load(
+        r#"
+        local stack_info = stack_info
+        local function baz(a, b, c, ...)
+            return stack_info()
+        end
+        assert(baz() == 'DebugStack { num_ups: 1, num_params: 3, is_vararg: true }')
+    "#,
+    )
+    .exec()?;
+
+    // LuaJIT does not pass this test for some reason
+    #[cfg(feature = "lua51")]
+    lua.load(
+        r#"
+        local stack_info = stack_info
+        local function baz(a, b, c, ...)
+            return stack_info()
+        end
+        assert(baz() == 'DebugStack { num_ups: 1 }')
+    "#,
+    )
+    .exec()?;
+
+    // Test retrieving currently running function
+    let running_function =
+        lua.create_function(|lua, ()| Ok(lua.inspect_stack(1, |debug| debug.function())))?;
+    lua.globals().set("running_function", running_function)?;
+    lua.load(
+        r#"
+        local function baz()
+            return running_function()
+        end
+        if jit == nil then
+            assert(baz() == baz)
+        else
+            -- luajit inline the "baz" function and returns the chunk itself
+            assert(baz() == running_function())
+        end
+    "#,
+    )
+    .exec()?;
+
+    Ok(())
+}
+
+#[test]
+fn test_traceback() -> Result<()> {
+    let lua = Lua::new();
+
+    // Test traceback at level 0 (not inside any function)
+    let traceback = lua.traceback(None, 0)?.to_string_lossy();
+    assert!(traceback.contains("stack traceback:"));
+
+    // Test traceback with a message prefix
+    let traceback = lua.traceback(Some("error occurred"), 0)?.to_string_lossy();
+    assert!(traceback.starts_with("error occurred"));
+    assert!(traceback.contains("stack traceback:"));
+
+    // Test traceback inside a function
+    let get_traceback = lua.create_function(|lua, (msg, level): (Option<StdString>, usize)| {
+        lua.traceback(msg.as_deref(), level)
+    })?;
+    lua.globals().set("get_traceback", get_traceback)?;
+
+    lua.load(
+        r#"
+        local function foo()
+            -- Level 1 is inside foo (the caller)
+            local traceback = get_traceback(nil, 1)
+            return traceback
+        end
+        local function bar()
+            local result = foo()
+            return result
+        end
+        local function baz()
+            local result = bar()
+            return result
+        end
+
+        local traceback = baz()
+        assert(traceback:match("in %a+ 'foo'"))
+        assert(traceback:match("in %a+ 'bar'"))
+        assert(traceback:match("in %a+ 'baz'"))
+    "#,
+    )
+    .exec()?;
+
+    // Test traceback at different levels
+    lua.load(
+        r#"
+        local function foo()
+            local tb0 = get_traceback(nil, 0)
+            local tb1 = get_traceback(nil, 1)
+            local tb2 = get_traceback(nil, 2)
+            return tb0, tb1, tb2
+        end
+        local function bar()
+            local tb0, tb1, tb2 = foo()
+            return tb0, tb1, tb2
+        end
+
+        local tb0, tb1, tb2 = bar()
+
+        assert(tb0:match("in %a+ 'get_traceback'"))
+        assert(tb0:match("in %a+ 'foo'"))
+
+        assert(not tb1:match("in %a+ 'get_traceback'"))
+        assert(tb1:match("in %a+ 'foo'"))
+
+        assert(not tb2:match("in %a+ 'foo'"))
+        assert(tb1:match("in %a+ 'bar'"))
+    "#,
+    )
+    .exec()?;
+
     Ok(())
 }
 
@@ -1240,7 +1470,7 @@ fn test_multi_states() -> Result<()> {
 
     let f = lua.create_function(|_, g: Option<Function>| {
         if let Some(g) = g {
-            g.call(())?;
+            g.call::<()>(())?;
         }
         Ok(())
     })?;
@@ -1289,10 +1519,16 @@ fn test_warnings() -> Result<()> {
     lua.set_warning_function(|_, _, _| Err(Error::runtime("warning error")));
     assert!(matches!(
         lua.load(r#"warn("test")"#).exec(),
-        Err(Error::CallbackError { cause, .. })
-            if matches!(*cause, Error::RuntimeError(ref err) if err == "warning error")
+        Err(Error::RuntimeError(ref err)) if err == "warning error"
     ));
 
+    // Recursive warning
+    lua.set_warning_function(|lua, _, _| {
+        lua.warning("inner", false);
+        Ok(())
+    });
+    lua.warning("hello", false);
+
     Ok(())
 }
 
@@ -1315,8 +1551,7 @@ fn test_luajit_cdata() -> Result<()> {
     "#,
         )
         .eval::<Value>()?;
-    assert!(cdata.is_userdata() && cdata.is_cdata());
-    assert_eq!(cdata.type_name(), "cdata");
+    assert_eq!(cdata.type_name(), "other");
     assert!(cdata.to_string()?.starts_with("cdata<void *>:"));
 
     Ok(())
@@ -1325,11 +1560,98 @@ fn test_luajit_cdata() -> Result<()> {
 #[test]
 #[cfg(feature = "send")]
 #[cfg(not(target_arch = "wasm32"))]
-fn test_send() {
+fn test_multi_thread() -> Result<()> {
     let lua = Lua::new();
-    std::thread::spawn(move || {
-        let _lua = lua;
-    })
-    .join()
-    .unwrap();
+
+    lua.globals().set("i", 0)?;
+    let func = lua.load("i = i + 1").into_function()?;
+
+    std::thread::scope(|s| {
+        s.spawn(|| {
+            for _ in 0..5 {
+                func.call::<()>(()).unwrap();
+            }
+        });
+        s.spawn(|| {
+            for _ in 0..5 {
+                func.call::<()>(()).unwrap();
+            }
+        });
+    });
+
+    assert_eq!(lua.globals().get::<i32>("i")?, 10);
+
+    Ok(())
+}
+
+#[test]
+fn test_exec_raw() -> Result<()> {
+    let lua = Lua::new();
+
+    let sum = lua.create_function(|_, args: Variadic<i32>| {
+        let mut sum = 0;
+        for i in args {
+            sum += i;
+        }
+        Ok(sum)
+    })?;
+    lua.globals().set("sum", sum)?;
+
+    let n: i32 = unsafe {
+        lua.exec_raw((), |state| {
+            ffi::lua_getglobal(state, b"sum\0".as_ptr() as _);
+            ffi::lua_pushinteger(state, 1);
+            ffi::lua_pushinteger(state, 7);
+            ffi::lua_call(state, 2, 1);
+        })
+    }?;
+    assert_eq!(n, 8);
+
+    // Test error handling
+    let res: Result<()> = unsafe {
+        lua.exec_raw("test error", |state| {
+            ffi::lua_error(state);
+        })
+    };
+    assert!(matches!(res, Err(Error::RuntimeError(err)) if err.contains("test error")));
+
+    Ok(())
+}
+
+#[test]
+fn test_gc_drop_ref_thread() -> Result<()> {
+    let lua = Lua::new();
+
+    let t = lua.create_table()?;
+    lua.create_function(move |_, ()| {
+        _ = &t;
+        Ok(())
+    })?;
+
+    for _ in 0..10000 {
+        // GC will run eventually to collect the function and the table above
+        lua.create_table()?;
+    }
+
+    Ok(())
+}
+
+#[cfg(not(feature = "luau"))]
+#[test]
+fn test_get_or_init_from_ptr() -> Result<()> {
+    // This would not work with Luau, the state must be init by mlua internally
+    let state = unsafe { ffi::luaL_newstate() };
+
+    let mut lua = unsafe { Lua::get_or_init_from_ptr(state) };
+    lua.globals().set("hello", "world678")?;
+
+    // The same Lua instance must be returned
+    lua = unsafe { Lua::get_or_init_from_ptr(state) };
+    assert_eq!(lua.globals().get::<String>("hello")?, "world678");
+
+    unsafe { ffi::lua_close(state) };
+
+    // Lua must not be accessed after closing
+
+    Ok(())
 }
